#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include "cx88_mpeg_ioctl.h"
#include "tuner_devnode_device.h"
#include "dtt7612.h"
#include "or51132.h"
#include "lgh064f.h"
#include "lg3303.h"
#include "mt2131.h"
#include "cx24227.h"

extern char *optarg;

void capture(uint32_t id, int fd, void *bufspace, cx88_mpeg_config &config, uint32_t time_s)
{
   cx88_mpeg_capture capture;
   capture.in_start_byte = CX88_MPEG_DEFAULT_START_BYTE;
   capture.in_start_byte_count = CX88_MPEG_DEFAULT_START_BYTE_COUNT;
   switch (id)
   {
      case 0x70633000:
         capture.in_options = 0;
         break;
      case 0x70635500:
         capture.in_options = CX88_MPEG_SERIAL_INPUT | CX88_MPEG_BYTEWIDE_SOP;
         break;
      case 0x00707801:
         capture.in_options = CX88_MPEG_SERIAL_INPUT | CX88_MPEG_PUNCTURED_CLOCK;
         break;
   }
   struct in_addr iaddr;
   struct sockaddr_in dest;
   socklen_t sz = 0;
   int outfile = 0;
   size_t msglen = (1500 / config.in_packet_size) * config.in_packet_size;
   if (!inet_aton("127.0.0.1", &iaddr))
   {
      printf("Error creating address\n");
      return;
   }
   if ((outfile = socket(PF_INET, SOCK_DGRAM, 0)) < 0)
   {
      printf("Error opening socket\n");
      return;
   }
   sz = sizeof(dest);
   if (getsockname(outfile, (struct sockaddr *)&dest, &sz) == -1)
   {
      printf("Error getting socket name\n");
      return;
   }
   dest.sin_port = ntohs(8802);
   dest.sin_addr = iaddr;
   if (connect(outfile, (struct sockaddr *)&dest, sz) == -1)
   {
      printf("Error connecting socket\n");
      return;
   }
   /*int outfile = open("out.ts", O_CREAT | O_APPEND | O_TRUNC | O_RDWR, 0777);
   if (outfile < 0)
   {
      printf("Error opening TS output file\n");
      return;  
   }*/
   time_t start_time = time(NULL);
   printf("Beginning capture\n");
   int error = ioctl(fd, CX88_MPEG_CAPTURE, &capture);
   if (error)
   {
      printf("Error during capture ioctl\n");
      return;
   }
   cx88_mpeg_wait wait;
   wait.in_timeout_s = 0;
   do
   {
      error = ioctl(fd, CX88_MPEG_WAIT, &wait);
      if (error)
      {
         printf("Error during wait ioctl\n");
         close(outfile);
         ioctl(fd, CX88_MPEG_STOP);
         return;
      }
      if ((wait.out_status & CX88_MPEG_BUS_ERROR) == CX88_MPEG_BUS_ERROR)
      {
         printf("Fatal error: PCI bus error\n");
      }
      if ((wait.out_status & CX88_MPEG_DMA_PROGRAM) == CX88_MPEG_DMA_PROGRAM)
      {
         printf("Fatal error: DMA program error\n");
      }
      if ((wait.out_status & CX88_MPEG_WAIT_ABORTED) == CX88_MPEG_WAIT_ABORTED)
      {
         printf("Fatal error: Wait aborted\n");
      }
      if ((wait.out_status & CX88_MPEG_CAP_TIMEOUT) == CX88_MPEG_CAP_TIMEOUT)
      {
         printf("Nonfatal error: Capture timed out\n");
      }
      if ((wait.out_status & CX88_MPEG_BAD_PACKET) == CX88_MPEG_BAD_PACKET)
      {
         printf("Nonfatal error: %d bad packets\n", wait.out_num_bad_packets);
      }
      if ((wait.out_status & CX88_MPEG_HW_OVERFLOW) == CX88_MPEG_HW_OVERFLOW)
      {
         printf("Nonfatal error: %d hardware buffer overflows\n", wait.out_num_overflows);
      }
      if (wait.out_status & CX88_MPEG_CAP_FATAL)
      {
         ioctl(fd, CX88_MPEG_STOP);
         close(outfile);
         return;
      }
      if (wait.out_buffer_available == 0 || wait.out_buffer_available > config.out_num_buffers)
      {
         printf("No buffer available\n");
         ioctl(fd, CX88_MPEG_STOP);
         close(outfile);
         return;
      }
      uint32_t bufoffset, bufsize;
      if (config.in_buftype == CX88_MPEG_BUF_KERNEL)
      {
         bufoffset = (wait.out_buffer_available - 1) * config.bufconfig.out_kernelbuf.out_buffer_size;
         bufsize = config.bufconfig.out_kernelbuf.out_mmap_size - bufoffset;
      }
      else
      {
         bufoffset = (wait.out_buffer_available - 1) * (config.out_packets_per_buffer * config.in_packet_size);
         bufsize = (config.inout_num_packets * config.in_packet_size) - bufoffset;
      }
      //write(outfile, (char *)bufspace + bufoffset, (bufsize > (config.out_packets_per_buffer * config.in_packet_size) ? (config.out_packets_per_buffer * config.in_packet_size) : bufsize));
      char *bufptr = (char *)bufspace + bufoffset;
      if (bufsize > (config.out_packets_per_buffer * config.in_packet_size))
      {
         bufsize = config.out_packets_per_buffer * config.in_packet_size;
      }
      uint32_t count = 0, writecount = 0;
      while (count < bufsize)
      {
         writecount = bufsize - count;
         if (writecount > msglen)
         {
            writecount = msglen;
         }
         ssize_t sz2 = write(outfile, bufptr, writecount);
         if (sz2 < 0)
         {
            break;
         }
         bufptr += writecount;
         count += writecount;
      }
   } while ((time(NULL) - start_time) < time_s);
   ioctl(fd, CX88_MPEG_STOP);
   close(outfile);
   printf("Packets transferred: %ld\n", wait.out_packet_count);
}

void tune(char *tuner_devnode, char *demod_devnode, uint32_t id, int fd, void *bufspace, cx88_mpeg_config &config, int time_s)
{
   int error = 0;
   dvb_channel channel;
   channel.modulation = DVB_MOD_VSB_8;
   channel.frequency_hz = 683000000;
   channel.bandwidth_hz = 6000000;
   channel.inversion = DVB_INVERSION_AUTO;
   tuner_config conf;
   error = conf.load_file("tuner_config");
   if (error)
   {
      printf("Error loading tuner config file\n");
      return;
   }
   tuner_devnode_device tuner_dev(conf, tuner_devnode, error);
   if (error)
   {
      printf("Error creating tuner devnode\n");
      return;
   }
   tuner_devnode_device demod_dev(conf, demod_devnode, error);
   if (error)
   {
      printf("Error creating demod devnode\n");
   }
   dvb_driver *tuner_driver = NULL, *demod_driver = NULL;
   switch (id)
   {
      case 0x70633000:
         tuner_driver = new dtt7612(conf, tuner_dev);
         demod_driver = new or51132(conf, demod_dev);
         break;
      case 0x70635500:
      {
         cx88_common_gpio gpio;
         gpio.in_write_mask = 0x1;
         gpio.in_read_mask = 0x0;
         gpio.inout_gpio = 0x0;
         if (ioctl(fd, CX88_COMMON_GPIO, &gpio) != 0)
         {
            printf("Error during GPIO ioctl\n");
            return;
         }
         usleep(100000);
         gpio.in_write_mask = 0x1;
         gpio.in_read_mask = 0x0;
         gpio.inout_gpio = 0x1;
         if (ioctl(fd, CX88_COMMON_GPIO, &gpio) != 0)
         {
            printf("Error during GPIO ioctl 2\n");
            return;  
         }
         usleep(200000);
         tuner_driver = new lgh064f(conf, tuner_dev);
         demod_driver = new lg3303(conf, demod_dev, DVB_IFC_NEG_POL, DVB_INPUT_SERIAL, error);
         break;
      }
      case 0x00707801:
         demod_driver = new cx24227(conf, demod_dev, DVB_INPUT_SERIAL, CX24227_QAM_IFREQ_44MHZ, CX24227_GPIO_ENABLE, CX24227_CLK_CONT_NINV, error);
         tuner_driver = new mt2131(conf, tuner_dev, error);
         break;
   }
   if (error || (tuner_driver == NULL) || (demod_driver == NULL))
   {
      printf("Error creating periphal drivers\n");
      goto cleanup;
   }
   dvb_interface ifc;
   tuner_driver->set_channel(channel, ifc);
   demod_driver->set_channel(channel, ifc);
   printf("Ready to start tuner\n");
   error = tuner_driver->start(10000);
   if (error)
   {
      printf("Error starting tuner\n");
      goto cleanup;
   }
   printf("Ready to start demod\n");
   error = demod_driver->start(2000);
   dvb_signal signal;
   demod_driver->get_signal(signal);
   printf("Signal stats = %g, %g\n", signal.snr, signal.strength);
   if (error)
   {
      printf("Error starting demod\n");
      tuner_driver->stop();
      goto cleanup;
   }
   capture(id, fd, bufspace, config, time_s);
cleanup:
   delete tuner_driver;
   delete demod_driver;
}

int main(int argc, char **argv)
{
   cx88_mpeg_config config;
   uint32_t id;
   uint32_t time_s = 0;
   int fd = open(argv[1], O_RDWR);
   int error = ioctl(fd, CX88_COMMON_CARD_ID, &id);
   int optch = -1;
   if (error)
   {
      printf("Error during card ID ioctl\n");
      close(fd);
      return 0;
   }
   printf("Card ID = 0x%x\n", id);
   void *buf = NULL;
   config.in_packet_size = CX88_MPEG_DEFAULT_PACKET_SIZE;
   config.in_buftype = CX88_MPEG_BUF_USER;  
   while ((optch = getopt(argc - 1, &(argv[1]), "u:k:t:")) != -1)
   {
      switch (optch)
      {
         case 't':
            time_s = strtoul(optarg, NULL, 10);
            break;
         case 'u':
         case 'k':
            if (buf != NULL)
            {
               printf("Buffer already allocated\n");
               goto cleanup;
            }
            config.inout_num_packets = strtoul(optarg, NULL, 10);
            if (optch == 'u')
            {
               config.in_buftype = CX88_MPEG_BUF_USER; 
               buf = malloc(config.inout_num_packets * config.in_packet_size);
               config.bufconfig.in_userbuf = buf;
            }
            else if (optch == 'k')
            {
               buf = MAP_FAILED;
               config.in_buftype = CX88_MPEG_BUF_KERNEL;
            }
            break;
      }
   }
   error = ioctl(fd, CX88_MPEG_CONFIG, &config);
   if (error)
   {
      printf("Error during config ioctl\n");
      goto cleanup;
   }
   printf("Actual num packets: %d\n", config.inout_num_packets);
   printf("Packets per buffer: %d\n", config.out_packets_per_buffer);
   if (config.in_buftype == CX88_MPEG_BUF_KERNEL)
   {
      buf = mmap(NULL, config.bufconfig.out_kernelbuf.out_mmap_size, PROT_READ, MAP_SHARED, fd, 0);
      if (buf == MAP_FAILED)
      {
         perror("Error mmap'ing device");
         goto cleanup;
      }
   }
   cx88_common_i2c_config i2config;
   i2config.in_bus_number = 0;
   i2config.in_num_i2c_devices = 2;
   sprintf(i2config.devices[0].in_driver, "iicdev");
   sprintf(i2config.devices[1].in_driver, "iicdev");
   switch (id)
   {
      case 0x70633000:
         i2config.devices[0].in_speed = CX88_I2C_SPEED_100KHZ;
         i2config.devices[0].in_addr = 0x61;
         i2config.devices[1].in_speed = CX88_I2C_SPEED_100KHZ;
         i2config.devices[1].in_addr = 0x15;
         break;
      case 0x70635500:
         i2config.devices[0].in_speed = CX88_I2C_SPEED_100KHZ;
         i2config.devices[0].in_addr = 0x61;
         i2config.devices[1].in_speed = CX88_I2C_SPEED_100KHZ;
         i2config.devices[1].in_addr = 0x59;
         break;
      case 0x00707801:
         i2config.devices[0].in_speed = CX88_I2C_SPEED_100KHZ;
         i2config.devices[0].in_addr = 0x61;
         i2config.devices[1].in_speed = CX88_I2C_SPEED_100KHZ;
         i2config.devices[1].in_addr = 0x19;
         break;
      default:
         printf("Unrecognized card type: %d\n", id);
         goto cleanup;
   }
   error = ioctl(fd, CX88_COMMON_I2C_CONFIG, &i2config);
   if (error)
   {
      printf("Error during I2C config ioctl\n");
      goto cleanup;  
   }
   char tuner_devnode[16], demod_devnode[16];
   sprintf(tuner_devnode, "/dev/iicdev%d", i2config.devices[0].out_instance);
   printf("Tuner: %s\n", tuner_devnode);
   sprintf(demod_devnode, "/dev/iicdev%d", i2config.devices[1].out_instance);
   printf("Demod: %s\n", demod_devnode);
   tune(tuner_devnode, demod_devnode, id, fd, buf, config, time_s);
cleanup:
   close(fd);
   if (config.in_buftype == CX88_MPEG_BUF_KERNEL)
   {
      if (buf != MAP_FAILED)
      {
         munmap(buf, config.bufconfig.out_kernelbuf.out_mmap_size);
      }
   }
   else
   {
      free(buf);
   }
   return 0;
}

